package luaAlchemy
{
	import events.ScriptResumeEvent;
	import components.ArgumentCollectorPanel;
	
	import flash.display.DisplayObject;
	import flash.events.Event;
	import flash.events.IOErrorEvent;
	import flash.events.SecurityErrorEvent;
	import flash.events.TimerEvent;
	import flash.filesystem.File;
	import flash.filesystem.FileMode;
	import flash.filesystem.FileStream;
	import flash.net.URLLoader;
	import flash.net.URLLoaderDataFormat;
	import flash.net.URLRequest;
	import flash.utils.Dictionary;
	import flash.utils.Timer;
	import flash.utils.getTimer;
	
	import luaAlchemy.LuaAlchemy;
	
	import mx.collections.ArrayCollection;
	import mx.collections.ArrayList;
	import mx.controls.Alert;
	import mx.events.CloseEvent;
	import mx.events.CollectionEvent;
	import mx.events.DynamicEvent;
	import mx.events.ListEvent;
	import mx.managers.PopUpManager;
	import mx.rpc.soap.WebService;
	
	import spark.components.Group;
	import spark.core.SpriteVisualElement;
	import spark.events.IndexChangeEvent;
	import spark.events.TextOperationEvent;
	
	/**
	 * A wrapper class for the lua script engine.
	 * 
	 */
	public class LuaCore extends SpriteVisualElement
	{
		/**
		 * The LuaAlchemy object.
		 */
		protected var lua:LuaAlchemy;
		
		/**
		 * The Script Wrapper object.
		 */
		protected var luaWrapper:String;
		
		/**
		 * The regular expression pattern that is searched for and replaced to insert scripts being executed.
		 * 
		 */
		protected var scriptPattern:RegExp = /--%%LUA-SCRIPT%%--/;
	
		/**
		 * The taskID prefix prepended to log entries.
		 * 
		 */
		protected var taskIDPrefix:String = "lua::task::";
		
		/**
		 * The current taskID index.
		 * 
		 */
		
		public var taskID:int = 0;
		
		/**
		 * A list of script starting times indexed by the running thread ID.
		 * 
		 */
		protected var scriptTimes:Object = {};
		
		/**
		 * The "view" object that this lua controller is embedded into.
		 * 
		 */
		protected var view:Object;
		
		/**
		 * A bindable collection that contains a list of log entries. Log entries are appended
		 * to the collection by using the printTrace() command from within lua.
		 * 
		 */
		[Bindable] public var logEntries:ArrayCollection = new ArrayCollection();
		
		/**
		 * The most recent result of the printTrace() command.
		 * 
		 */
		[Bindable] public var lastLogText:String;
		
		/**
		 * The ArgumentCollectorPanel used by scripts that need to collect information.
		 */
		protected var ac:ArgumentCollectorPanel;
		
		
		public function LuaCore( )
		{
			super();
			logEntries.addEventListener(CollectionEvent.COLLECTION_CHANGE, onLogChange); 
		}
		
		/**
		 * Initializes the lua scripting engine.
		 * 
		 * @param targetView A component that is set into a lua global variable named "view".
		 * @param libraries An array of paths to some lua libraries that will be loaded into the scripting engine.
		 * 
		 */
		public function init( targetView:Object, libraries:Array ) : void {
			
			if (targetView == null)
				return;
			
			targetView = view;
			var errorText:String;
			lua = new LuaAlchemy();
			lua.setGlobal("view", view );
			lua.setGlobal("modelCore", this );
			
			// var systemLib:File;
			var fault:Boolean = false;
			
			for each (var a:String in libraries) {
				fault = loadLuaLib( a ) || fault;
			}
			
			fault == false ? trace('System fault occurred') : void;
			
			var systemLib:File = File.applicationDirectory.resolvePath("runtime_assets/Scripts/wrapper.lua"); 
			
			var stream:FileStream = new FileStream();
			stream.open( systemLib, FileMode.READ );
			luaWrapper = stream.readUTFBytes( stream.bytesAvailable );
			
		}
		
		/**
		 * Load and execute into the lua core the provided system library.
		 * 
		 * @param libPath The path of the target library.
		 */
		public function loadLuaLib( libPath:String ) : Boolean {
			var systemLib:File;
			systemLib = File.applicationDirectory.resolvePath(libPath); // This should work regardless of whether it's an absolute system path or relative 
			var fault:Boolean = false;
			var stream:FileStream = new FileStream();
			stream.open( systemLib, FileMode.READ );
			var sLib:String = stream.readUTFBytes( stream.bytesAvailable );
			var stack:Array = lua.doString( sLib );
			if (stack.shift == false) {
				logEntries.addAll( new ArrayList( stack ));//stack.join("\n") + "\n";	
				mx.controls.Alert.show("Warning: The LUA Controller was unable to load the system library: \n " + libPath, "Warning: Unable to load library!" );
				return false;
			}
			return true;
		}
		
		
		/**
		 * Loads and executes a .lua script file from the supplied path.
		 * 
		 * @param script The path of the file to load.
		 */
		public function executeTask( script:String ) : void {
			lua.setGlobal("taskID", taskIDPrefix + taskID);				
			scriptTimes[ taskIDPrefix + taskID ] = flash.utils.getTimer();
			var scriptFile:File;
			scriptFile = File.applicationDirectory.resolvePath(script); // This should work regardless of whether it's an absolute system path or relative 
			var fault:Boolean = false;
			var stream:FileStream = new FileStream();
			stream.open( scriptFile, FileMode.READ );
			var scriptTxt:String = stream.readUTFBytes( stream.bytesAvailable );
			var stack:Array = lua.doString( luaWrapper.replace( scriptPattern, scriptTxt ));
			taskID++;
		}
		
		/**
		 * Runs a bare script provided via the script variable.
		 */
		public function eval( script:String ) : void {
			lua.setGlobal("taskID", taskIDPrefix + taskID);				
			scriptTimes[ taskIDPrefix + taskID ] = flash.utils.getTimer();
			var stack:Array = lua.doString( luaWrapper.replace( scriptPattern, script ));
			// trace(luaWrapper.replace( scriptPattern, script ));
			taskID++;
		}		

		/**
		 * taskComplete is called from within the lua controller via the "modelCore" reference.
		 * 
		 * @param threadID The ID of the running coroutine task.
		 * 
		 */
		
		public function taskComplete( threadID:String ) : void {
			logEntries.addItem( "Task : " + threadID + " complete in " + (flash.utils.getTimer() - scriptTimes[ threadID ]) + " ms");
			delete scriptTimes[ threadID ];
		}
		
		/**
		 * taskFailed is called from within the lua controller via the "modelCore" reference.
		 * 
		 * @param threadID The ID of the running coroutine task.
		 * 
		 */
		public function taskFailed( threadID:String ) : void {
			logEntries.addItem( "Task : " + threadID + " failed.in " + (flash.utils.getTimer() - scriptTimes[ threadID ]) + " ms");
			delete scriptTimes[ threadID ];
		}
		
		/**
		 * Sets a variable global variable in the Lua environment.
		 * 
		 * @param key The name of the global variable to set.
		 * @param value The name of the value to assign to the new variable.
		 * @public
		 */
		public function setCoreScopeValue( key:String, value:Object ) : void {
			lua.setGlobal( key, value );
		}
		
		/**
		 * Handle an update to the log.
		 * 
		 */
		protected function onLogChange( e:CollectionEvent ) : void {
			if (logEntries.length < 1)
				return;
			lastLogText = logEntries.getItemAt( logEntries.length - 1 ) as String;
		}
		
		/**
		 * OPens a window that prompts the user for information.
		 *
		 * @param threadID The id of the running coroutine.
		 * @param message The message prompt for the argument collector.
		 * @param arguments The list of arguments to ask for. 
		 */
		public function promptForArguments( threadID:String, message:String, arguments:Array ) : Boolean {
			var popupTarget:DisplayObject = view == null ? this : view as DisplayObject;
			ac = PopUpManager.createPopUp( popupTarget, ArgumentCollectorPanel, true) as ArgumentCollectorPanel;
			ac.setArguments( threadID, { arguments : arguments, prompt : message, __taskID__ : threadID });
			ac.addEventListener( "argumentsResponse", onArguments);
			ac.addEventListener( "closeView", onAcClose );
			PopUpManager.centerPopUp( ac );
			return true;
		}
		
		/**
		 * The "Cancel handler" for the ArgumentsCollector panel. This is fired on cancelling the form.
		 * 
		 * @param e The CloseEvent.
		 */
		protected function onAcClose( e:DynamicEvent ) : void {
			PopUpManager.removePopUp( ac );
			var se:ScriptResumeEvent = new ScriptResumeEvent( e.taskID, 1, e.results );
			dispatchEvent( se );
		}
		
		/**
		 * The "onArguments" handler for the ArgumentsCollector panel. This is fired on clicking "OK" on the form.
		 * 
		 * @param e The onArguments event.
		 */
		public function onArguments( e:DynamicEvent ) : void {
			PopUpManager.removePopUp( ac );
			var taskID:String = e.taskID;
			var results:Object = e.results;
			var se:ScriptResumeEvent = new ScriptResumeEvent( taskID, 0, results);
			dispatchEvent( se );
		}
		
		/**
		 * A list of alert controls, indexed by thread ID of the running tasks that called the alert via askYesOrNo.
		 * 
		 */
		private var alertTasks:Dictionary = new Dictionary();
		
		/**
		 * Script-callable function to pop-up a yes/no dialog for asking simple questions.
		 * 
		 */
		public function askYesOrNo( taskID:String, message:String ) : void {
			var alert:Alert = Alert.show( message, "Prompt", Alert.YES | Alert.NO, null, handleYesNo );
			alertTasks[ alert ] = taskID;
		}
		
		/**
		 * Adds an entry to the logEntries collection.
		 * 
		 */
		public function logMessage( message:String ) : void {
			logEntries.addItem( message );
		}
		
		/**
		 * Handler for the yes/no function.
		 */
		public function handleYesNo( e:CloseEvent ) : void {
			var taskID:String = alertTasks[ e.target ];
			var r:String = e.detail == Alert.YES ? "yes" : "no";
			delete alertTasks[ e.target ];
			var se:ScriptResumeEvent = new ScriptResumeEvent( taskID, 0, r );
			dispatchEvent( se );
			
		} 
		
		/**
		 * Returns the result of flash.utils.getTimer()
		 * 
		 */
		public function getTimerStep() : int {	
			var val:int = flash.utils.getTimer();
			return val;
		}
		
		/**
		 * A list of Timer objects, indexed by the taskID.
		 */
		private var sleepingTasks:Dictionary = new Dictionary();
		/**
		 * A vector to contain the timers, probably redundant.
		 * 
		 */
		private var timers:Vector.<Timer> = new Vector.<Timer>;
		
		/**
		 * Creates a single-shot timer to supply a "sleep" functionality for the lua core.
		 */
		public function sleep( taskID:String, time:int ) : void {
			var sleepTimer:Timer = new Timer( time, 1);
			sleepTimer.addEventListener(TimerEvent.TIMER_COMPLETE, sleepTimerResume, false, 0, true );
			timers.push( sleepTimer );
			sleepTimer.start();
			sleepingTasks[ sleepTimer ] = taskID;
		}
		
		/**
		 * Resumes a running thread when the sleep timer expires.
		 * 
		 */
		public function sleepTimerResume( e:TimerEvent ) : void {
			var taskID:String = sleepingTasks[ e.target ];
			delete sleepingTasks[ e.target ];
			timers.splice( timers.indexOf( e.target ), 1 );
			var se:ScriptResumeEvent = new ScriptResumeEvent( taskID, 0, null );
			dispatchEvent( se );
		}
		
		/**
		 * A list of URLLoaders that are currently loading data, probably redundant.
		 */
		private var loaders:Vector.<URLLoader> = new Vector.<URLLoader>;
		
		/**
		 * A list of loader objects, indexed by taskID. 
		 * 
		 */
		private var loadingTasks:Dictionary = new Dictionary();
		
		/**
		 * Loads data from a URL passed in from the script. 
		 * 
		 * @param taskID The ThreadID of the task that called the operation.
		 * @param url The URL to load.
		 * @param dataType The kind of data to load. If dataType is "binary" then the urlLoader's dataFormat parameter is set accordingly.
		 * 
		 */
		public function loadDataFromURL( taskID:String, url:String, dataFormat:String = "text") : void {
			
			var loader:URLLoader = new URLLoader();
			if (dataFormat == "binary") {
				loader.dataFormat = URLLoaderDataFormat.BINARY;
			}
			loaders.push( loader );
			loadingTasks[ loader ] = taskID;
			loader.addEventListener(Event.COMPLETE, urlLoader_completeHandler, false, 0, true );
			loader.addEventListener(SecurityErrorEvent.SECURITY_ERROR, urlLoader_securityError, false, 0, true );
			loader.addEventListener(IOErrorEvent.IO_ERROR, urlLoader_ioErrorHandler, false, 0, true );
			loader.load( new URLRequest( url ));
		}
		
		/**
		 * Complete handler for the URLLoader.
		 * @param e The complete event.
		 */
		private function urlLoader_completeHandler( e:Event ) : void {
			var taskID:String = loadingTasks[ e.target ];
			loaders.splice( loaders.indexOf( e.target ), 1);
			var se:ScriptResumeEvent = new ScriptResumeEvent( taskID, 0, e.target.data );
			dispatchEvent( se );
		}
		
		/**
		 * Security error handler for the URLLoader.
		 * @param e The security error event.
		 * 
		 * Dispatches a ScriptResumeEvent with the error parameter set to 1.
		 */
		private function urlLoader_securityError( e:Event ) : void {
			var taskID:String = loadingTasks[ e.target ];
			loaders.splice( loaders.indexOf( e.target ), 1);
			var se:ScriptResumeEvent = new ScriptResumeEvent( taskID, 1, "securityError");
			dispatchEvent( se );
		}
		
		/**
		 * IOError handler for the URLLoader.
		 * @param e The IOError event.
		 * 
		 * Dispatches a ScriptResumeEvent with the error parameter set to 1.
		 * 
		 */
		private function urlLoader_ioErrorHandler( e:Event ) : void {
			var taskID:String = loadingTasks[ e.target ];
			loaders.splice( loaders.indexOf( e.target ), 1);
			var se:ScriptResumeEvent = new ScriptResumeEvent( taskID, 0, "ioError" );
			dispatchEvent( se );
		}
		
	}
}