-- Begin lua core coroutine wrapper
local run = coroutine.create( function( )
	
	-- The threadTaskID is assigned from the global taskID identifier because THIS coroutine will notify
	-- the containing component with its assigned ID that this task was resumed and ended successfully in the case of a
	-- coroutine.yield() call.
	local threadTaskID = as3.tolua(taskID)

	-- This is the errorCode that will be processed after the script block is run.
	local errorCode = 0

	-- Add this coroutine to this lua engine's environment using threadTaskID 
	waiting_threads[ threadTaskID ] = coroutine.running()

	-- Set the running_thread variable so functions calls that communicate with the application core
	-- can provide it as a reference
	running_thread = threadTaskID
		
	-- LUA script block follows
	
	--%%LUA-SCRIPT%%--

	-- LUA script block precedes
	
	-- Remove this task from the list of threads, because it's now complete
	-- We use threadTaskID here since the global value taskID will be different if a coroutine
	-- yielded control (e.g. a script called waitForResponse )
	waiting_threads[ threadTaskID ] = nil
	
	-- Clear the target var. The script will be able to access it via targets[ threadTaskID ]
	target = nil
	
	if (errorCode == 0) then
		modelCore.taskComplete( threadTaskID ) -- Notify the As3 component containing the lua engine that this script is complete
	else
		modelCore.taskFailed( threadTaskID ) --- Notify the As3 component containing the lua engine that this script failed.
	end
	-- return threadTaskID
end)

assert(coroutine.resume( run ))
-- End LUA core coroutine wrapper